@@ -24,8 +24,6 @@ typedef struct Manager Manager;
 
 /// Additional includes needed by elogind
 #include "cgroup-util.h"
-#include "elogind.h"
-#include "musl_missing.h"
 #include "sleep-config.h"
 
 #if 1 /// elogind has to ident itself
@@ -82,6 +80,13 @@ struct Manager {
         /* Data specific to the cgroup subsystem */
         CGroupMask cgroup_supported;
         char *cgroup_root;
+
+        /* elogind might spawn processes to suspend/hibernate, so we need their PIDs to end them properly */
+        pid_t sleep_fork_pid; /* for suspend/hibernate fork */
+        pid_t tool_fork_pid;  /* for external tool fork */
+
+        /* To wake up sleeping consumers using the right operation, the manager must know what is going on. */
+        const HandleActionData *sleep_fork_action;
 #endif // 0
 
         Seat *seat0;
@@ -89,8 +94,8 @@ struct Manager {
         char **kill_only_users, **kill_exclude_users;
         bool kill_user_processes;
 
-        unsigned long session_counter;
-        unsigned long inhibit_counter;
+        uint64_t session_counter;
+        uint64_t inhibit_counter;
 
 #if 0 /// elogind does not support session units
         Hashmap *session_units;
@@ -108,13 +113,12 @@ struct Manager {
         /* If a shutdown/suspend is currently executed, then this is the job of it */
         char *action_job;
 #else // 0
-        /* Suspension and hibernation can be disabled in logind.conf. */
+        /* Suspension and hibernation can be disabled in sleep.conf. */
         bool allow[_SLEEP_OPERATION_MAX];
         char **modes[_SLEEP_OPERATION_MAX];
         char **states[_SLEEP_OPERATION_MAX];
-        usec_t hibernate_delay_usec; /* HibernateDelaySec */
-        usec_t suspend_estimation_usec; /* SuspendEstimationSec */
-
+        usec_t hibernate_delay_usec;
+        usec_t suspend_estimation_usec;
 
         /* If an admin puts scripts into SYSTEM_SLEEP_PATH and/or
            SYSTEM_POWEROFF_PATH that fail, the ongoing suspend/poweroff
@@ -225,7 +229,7 @@ bool manager_shall_kill(Manager *m, const char *user);
 int manager_get_idle_hint(Manager *m, dual_timestamp *t);
 
 int manager_get_user_by_pid(Manager *m, pid_t pid, User **user);
-int manager_get_session_by_pid(Manager *m, pid_t pid, Session **session);
+int manager_get_session_by_pidref(Manager *m, const PidRef *pid, Session **ret);
 
 bool manager_is_lid_closed(Manager *m);
 bool manager_is_docked_or_external_displays(Manager *m);
