@@ -13,5 +13,7 @@ int read_reboot_parameter(char **parameter);
 #if 0 /// UNNEEDED by elogind
 int reboot_with_parameter(RebootFlags flags);
 
-int shall_restore_state(void);
 #endif // 0
+bool shall_restore_state(void);
+
+bool kexec_loaded(void);
