@@ -5,9 +5,10 @@
 
 #include "sd-daemon.h"
 
+#include "argv-util.h"
 #include "macro.h"
 #include "static-destruct.h"
-#include "util.h"
+#include "strv.h"
 
 #if 0 /// UNNEEDED by elogind
 static inline bool manager_errno_skip_test(int r) {
@@ -47,7 +48,9 @@ void test_setup_logging(int level);
                 EXIT_TEST_SKIP;                                         \
         })
 
+#if 0 /// UNNEEDED by elogind
 int write_tmpfile(char *pattern, const char *contents);
+#endif // 0
 
 bool have_namespaces(void);
 
@@ -57,6 +60,12 @@ bool have_namespaces(void);
 bool can_memlock(void);
 #endif // 0
 
+/* Define void* buffer and size_t length variables from a hex string. */
+#define DEFINE_HEX_PTR(name, hex)                                       \
+        _cleanup_free_ void *name = NULL;                               \
+        size_t name##_len = 0;                                          \
+        assert_se(unhexmem(hex, strlen_ptr(hex), &name, &name##_len) >= 0);
+
 #define TEST_REQ_RUNNING_SYSTEMD(x)                                 \
         if (sd_booted() > 0) {                                      \
                 x;                                                  \
@@ -103,14 +112,31 @@ extern const TestFunc _weak_ __stop_SYSTEMD_TEST_TABLE[];
         REGISTER_TEST(test_##name, ##__VA_ARGS__); \
         static int test_##name(void)
 
+#define TEST_LOG_FUNC() \
+        log_info("/* %s() */", __func__)
+
 static inline int run_test_table(void) {
+        _cleanup_strv_free_ char **tests = NULL;
         int r = EXIT_SUCCESS;
+        bool ran = false;
+        const char *e;
 
         if (!__start_SYSTEMD_TEST_TABLE)
                 return r;
 
-        const TestFunc *t = ALIGN_PTR(__start_SYSTEMD_TEST_TABLE);
-        while (t < __stop_SYSTEMD_TEST_TABLE) {
+        e = getenv("TESTFUNCS");
+        if (e) {
+                r = strv_split_full(&tests, e, ":", EXTRACT_DONT_COALESCE_SEPARATORS);
+                if (r < 0)
+                        return log_error_errno(r, "Failed to parse $TESTFUNCS: %m");
+        }
+
+        for (const TestFunc *t = ALIGN_PTR(__start_SYSTEMD_TEST_TABLE);
+             t + 1 <= __stop_SYSTEMD_TEST_TABLE;
+             t = ALIGN_PTR(t + 1)) {
+
+                if (tests && !strv_contains(tests, t->name))
+                        continue;
 
                 if (t->sd_booted && sd_booted() <= 0) {
                         log_info("/* elogind not booted, skipping %s */", t->name);
@@ -127,9 +153,12 @@ static inline int run_test_table(void) {
                                 t->f.void_func();
                 }
 
-                t = ALIGN_PTR(t + 1);
+                ran = true;
         }
 
+        if (!ran)
+                return log_error_errno(SYNTHETIC_ERRNO(ENXIO), "No matching tests found.");
+
         return r;
 }
 
