@@ -6,38 +6,46 @@
 
 #include <errno.h>
 #include <fcntl.h>
-#include <getopt.h>
-//#include <linux/fiemap.h>
+//#include <getopt.h>
 #include <poll.h>
-#include <sys/stat.h>
-//#include <sys/types.h>
 #include <sys/timerfd.h>
+//#include <sys/types.h>
+#include <sys/utsname.h>
 #include <unistd.h>
 
 #include "sd-bus.h"
+#include "sd-device.h"
+#include "sd-id128.h"
 #include "sd-messages.h"
 
-//#include "btrfs-util.h"
+#include "battery-capacity.h"
+#include "battery-util.h"
+//#include "build.h"
 //#include "bus-error.h"
 //#include "bus-locator.h"
 //#include "bus-util.h"
-#include "def.h"
+#include "constants.h"
+//#include "devnum-util.h"
+#include "efivars.h"
 #include "exec-util.h"
 #include "fd-util.h"
 #include "fileio.h"
 //#include "format-util.h"
+#include "hibernate-util.h"
+#include "id128-util.h"
 #include "io-util.h"
+#include "json.h"
 #include "log.h"
 //#include "main-func.h"
+#include "os-util.h"
 //#include "parse-util.h"
 //#include "pretty-print.h"
 #include "sleep-config.h"
 //#include "special.h"
-#include "stdio-util.h"
+//#include "stdio-util.h"
 #include "string-util.h"
 #include "strv.h"
 #include "time-util.h"
-//#include "util.h"
 
 /// Additional includes needed by elogind
 #include <stdio.h>
@@ -46,35 +54,36 @@
 #include "sd-login.h"
 #include "sleep.h"
 #include "terminal-util.h"
-#include "utmp-wtmp.h"
+#include "wall.h"
+
 #define DEFAULT_HIBERNATE_DELAY_USEC_NO_BATTERY (2 * USEC_PER_HOUR)
 
 static SleepOperation arg_operation = _SLEEP_OPERATION_INVALID;
 
 #if 1 /// If an nvidia card is present, elogind informs its driver about suspend/resume actions
-static int nvidia_sleep(Manager* m, SleepOperation operation, unsigned* vtnr) {
+static int nvidia_sleep(Manager* m, int fd, SleepOperation operation, unsigned* vtnr) {
         static char const* drv_suspend = "/proc/driver/nvidia/suspend";
-        int r;
+        _cleanup_fclose_ FILE *f = NULL;
+        int k, r;
         char** sessions;
-        struct stat std;
         unsigned vt = 0;
 
         assert(m);
+        assert(fd >= 0);
         assert(operation >= 0);
         assert(operation <= _SLEEP_OPERATION_MAX);
         assert(operation != SLEEP_SUSPEND_THEN_HIBERNATE); /* execute_s2h() calls execute() with suspend/hibernate */
         assert(vtnr);
 
-        // See whether an nvidia suspension is possible
-        r = stat(drv_suspend, &std);
-        if (r)
-                return 0;
+        k = fdopen_independent(fd, "we", &f);
+        if (k < 0)
+                return RET_GATHER(r, k);
 
         if (operation != _SLEEP_OPERATION_MAX) {
                 *vtnr = 0;
 
                 // Find the (active) sessions of the sleep sender
-                r = sd_uid_get_sessions(m->scheduled_sleep_uid, 1, &sessions);
+                r = sd_uid_get_sessions(m ? m->scheduled_sleep_uid : 0, 1, &sessions);
 #if ENABLE_DEBUG_ELOGIND
                 char *t = strv_join(sessions, " ");
                 log_debug_elogind("sd_uid_get_sessions() returned %d, result is: %s", r, strnull(t));
@@ -85,9 +94,8 @@ static int nvidia_sleep(Manager* m, SleepOperation operation, unsigned* vtnr) {
 
                 // Find one with a VT (Really, sessions should hold only one active session anyway!)
                 STRV_FOREACH(session, sessions) {
-                        int k;
-                        k = sd_session_get_vt(*session, &vt);
-                        if (k >= 0) {
+                        int s = sd_session_get_vt(*session, &vt);
+                        if (s >= 0) {
                                 log_debug_elogind("Active session %s with VT %u found", *session, vt);
                                 break;
                         }
@@ -103,20 +111,33 @@ static int nvidia_sleep(Manager* m, SleepOperation operation, unsigned* vtnr) {
                 }
 
                 // Okay, go to sleep.
+
                 if (operation == SLEEP_SUSPEND) {
                         log_debug_elogind("Writing 'suspend' into %s", drv_suspend);
-                        r = write_string_file(drv_suspend, "suspend", WRITE_STRING_FILE_DISABLE_BUFFER);
+                        k = write_string_stream(f, "suspend", WRITE_STRING_FILE_DISABLE_BUFFER);
+                        if (k >= 0) {
+                                return 1;
+                        }
+                        RET_GATHER(r, log_debug_errno(k, "Failed to write '%s' to %s: %m", "suspend", drv_suspend));
                 } else {
                         log_debug_elogind("Writing 'hibernate' into %s", drv_suspend);
-                        r = write_string_file(drv_suspend, "hibernate", WRITE_STRING_FILE_DISABLE_BUFFER);
+                        k = write_string_stream(f, "hibernate", WRITE_STRING_FILE_DISABLE_BUFFER);
+                        if (k >= 0) {
+                                return 1;
+                        }
+                        RET_GATHER(r, log_debug_errno(k, "Failed to write '%s' to %s: %m", "hibernate", drv_suspend));
                 }
 
-                if (r)
-                        return 0;
+                return 0;
         } else {
                 // Wakeup the device
                 log_debug_elogind("Writing 'resume' into %s", drv_suspend);
-                (void) write_string_file(drv_suspend, "resume", WRITE_STRING_FILE_DISABLE_BUFFER);
+                k = write_string_stream(f, "resume", WRITE_STRING_FILE_DISABLE_BUFFER);
+                if (k >= 0) {
+                        return 1;
+                }
+                RET_GATHER(r, log_debug_errno(k, "Failed to write '%s' to %s: %m", "resume", drv_suspend));
+
                 // Then try to change back
                 if (*vtnr > 0) {
                         log_debug_elogind("Switching back to VT %u", *vtnr);
@@ -133,31 +154,33 @@ static int execute_external(Manager *m, SleepOperation operation) {
         int r = -1;
         char **tools = NULL;
 
-        if ( (SLEEP_SUSPEND == operation) && !strv_isempty(m->suspend_by_using) )
+        if ((SLEEP_SUSPEND == operation) && !strv_isempty(m->suspend_by_using))
                 tools = m->suspend_by_using;
 
-        if ( ( (SLEEP_HIBERNATE == operation) || (SLEEP_HYBRID_SLEEP == operation) ) && !strv_isempty(m->hibernate_by_using) )
+        if (((SLEEP_HIBERNATE == operation) || (SLEEP_HYBRID_SLEEP == operation)) && !strv_isempty(m->hibernate_by_using))
                 tools = m->hibernate_by_using;
 
-        if ( tools ) {
+        if (tools) {
                 STRV_FOREACH(tool, tools) {
                         int k;
 
                         log_debug_elogind("Calling '%s' to '%s'...", *tool, sleep_operation_to_string(operation));
-                        k = safe_fork( *tool, FORK_RESET_SIGNALS | FORK_REOPEN_LOG, NULL );
+                        k = safe_fork(*tool, FORK_LOG|FORK_RESET_SIGNALS|FORK_CLOSE_ALL_FDS, &m->tool_fork_pid);
 
-                        if ( k < 0 ) {
+                        if (k < 0) {
                                 r = log_error_errno(errno, "Failed to fork run %s: %m", *tool);
                                 continue;
                         }
 
-                        if ( 0 == k ) {
+                        if (0 == k) {
                                 /* Child */
-                                execlp( *tool, *tool, NULL );
-                                log_error_errno( errno, "Failed to execute %s: %m", *tool );
-                                _exit( EXIT_FAILURE );
+                                execlp(*tool, *tool, NULL);
+                                log_error_errno(errno, "Failed to execute %s: %m", *tool);
+                                _exit(EXIT_FAILURE);
                         }
 
+                        log_debug_elogind( "Forking off '%s' returned %d and set PID %d", *tool, k, m->tool_fork_pid );
+
                         return 0;
                 }
         }
@@ -166,90 +189,125 @@ static int execute_external(Manager *m, SleepOperation operation) {
 }
 #endif // 1
 
-static int write_hibernate_location_info(const HibernateLocation *hibernate_location) {
-        char offset_str[DECIMAL_STR_MAX(uint64_t)];
-        char resume_str[DECIMAL_STR_MAX(unsigned) * 2 + STRLEN(":")];
-        int r;
-#if 1 /// To support LVM setups, elogind uses device numbers
-        char device_num_str[DECIMAL_STR_MAX(uint32_t) * 2 + 2];
-        struct stat stb;
-#endif // 1
+static int write_efi_hibernate_location(const HibernationDevice *hibernation_device, bool required) {
+        int log_level = required ? LOG_ERR : LOG_DEBUG;
 
-        assert(hibernate_location);
-        assert(hibernate_location->swap);
+#if ENABLE_EFI
+        _cleanup_(json_variant_unrefp) JsonVariant *v = NULL;
+        _cleanup_free_ char *formatted = NULL, *id = NULL, *image_id = NULL,
+                       *version_id = NULL, *image_version = NULL;
+        _cleanup_(sd_device_unrefp) sd_device *device = NULL;
+        const char *uuid_str;
+        sd_id128_t uuid;
+        struct utsname uts = {};
+        int r, log_level_ignore = required ? LOG_WARNING : LOG_DEBUG;
 
-        xsprintf(resume_str, "%u:%u", major(hibernate_location->devno), minor(hibernate_location->devno));
-        r = write_string_file("/sys/power/resume", resume_str, WRITE_STRING_FILE_DISABLE_BUFFER);
+        assert(hibernation_device);
 
-#if 1 /// To support LVM setups, elogind uses device numbers if the direct approach failed
-        if (r < 0) {
-                r = stat(hibernate_location->swap->device, &stb);
-                if (r < 0)
-                        return log_debug_errno(errno, "Error while trying to get stats for %s: %m", hibernate_location->swap->device);
-
-                (void) snprintf(device_num_str, DECIMAL_STR_MAX(uint32_t) * 2 + 2,
-                                "%u:%u",
-                                major(stb.st_rdev), minor(stb.st_rdev));
-                r = write_string_file("/sys/power/resume", device_num_str, 0);
-        }
-#endif // 1
+        if (!is_efi_boot())
+                return log_full_errno(log_level, SYNTHETIC_ERRNO(EOPNOTSUPP),
+                                      "Not an EFI boot, passing HibernateLocation via EFI variable is not possible.");
 
+        r = sd_device_new_from_devnum(&device, 'b', hibernation_device->devno);
         if (r < 0)
-                return log_debug_errno(r, "Failed to write partition device to /sys/power/resume for '%s': '%s': %m",
-                                       hibernate_location->swap->device, resume_str);
+                return log_full_errno(log_level, r, "Failed to create sd-device object for '%s': %m",
+                                      hibernation_device->path);
 
-        log_debug("Wrote resume= value for %s to /sys/power/resume: %s", hibernate_location->swap->device, resume_str);
+        r = sd_device_get_property_value(device, "ID_FS_UUID", &uuid_str);
+        if (r < 0)
+                return log_full_errno(log_level, r, "Failed to get filesystem UUID for device '%s': %m",
+                                      hibernation_device->path);
 
-        /* if it's a swap partition, we're done */
-        if (streq(hibernate_location->swap->type, "partition"))
-                return r;
+        r = sd_id128_from_string(uuid_str, &uuid);
+        if (r < 0)
+                return log_full_errno(log_level, r, "Failed to parse ID_FS_UUID '%s' for device '%s': %m",
+                                      uuid_str, hibernation_device->path);
 
-        if (!streq(hibernate_location->swap->type, "file"))
-                return log_debug_errno(SYNTHETIC_ERRNO(EINVAL),
-                                       "Invalid hibernate type: %s", hibernate_location->swap->type);
+        if (uname(&uts) < 0)
+                log_full_errno(log_level_ignore, errno, "Failed to get kernel info, ignoring: %m");
 
-        /* Only available in 4.17+ */
-        if (hibernate_location->offset > 0 && access("/sys/power/resume_offset", W_OK) < 0) {
-                if (errno == ENOENT) {
-                        log_debug("Kernel too old, can't configure resume_offset for %s, ignoring: %" PRIu64,
-                                  hibernate_location->swap->device, hibernate_location->offset);
+        r = parse_os_release(NULL,
+                             "ID", &id,
+                             "IMAGE_ID", &image_id,
+                             "VERSION_ID", &version_id,
+                             "IMAGE_VERSION", &image_version);
+        if (r < 0)
+                log_full_errno(log_level_ignore, r, "Failed to parse os-release, ignoring: %m");
+
+        r = json_build(&v, JSON_BUILD_OBJECT(
+                               JSON_BUILD_PAIR_UUID("uuid", uuid),
+                               JSON_BUILD_PAIR_UNSIGNED("offset", hibernation_device->offset),
+                               JSON_BUILD_PAIR_CONDITION(!isempty(uts.release), "kernelVersion", JSON_BUILD_STRING(uts.release)),
+                               JSON_BUILD_PAIR_CONDITION(id, "osReleaseId", JSON_BUILD_STRING(id)),
+                               JSON_BUILD_PAIR_CONDITION(image_id, "osReleaseImageId", JSON_BUILD_STRING(image_id)),
+                               JSON_BUILD_PAIR_CONDITION(version_id, "osReleaseVersionId", JSON_BUILD_STRING(version_id)),
+                               JSON_BUILD_PAIR_CONDITION(image_version, "osReleaseImageVersion", JSON_BUILD_STRING(image_version))));
+        if (r < 0)
+                return log_full_errno(log_level, r, "Failed to build JSON object: %m");
+
+        r = json_variant_format(v, 0, &formatted);
+        if (r < 0)
+                return log_full_errno(log_level, r, "Failed to format JSON object: %m");
+
+        r = efi_set_variable_string(EFI_SYSTEMD_VARIABLE(HibernateLocation), formatted);
+        if (r < 0)
+                return log_full_errno(log_level, r, "Failed to set EFI variable HibernateLocation: %m");
+
+        log_debug("Set EFI variable HibernateLocation to '%s'.", formatted);
+        return 0;
+#else
+        return log_full_errno(log_level, SYNTHETIC_ERRNO(EOPNOTSUPP),
+                              "EFI support not enabled, passing HibernateLocation via EFI variable is not possible.");
+#endif
+}
+
+static int write_state(int fd, char * const *states) {
+        int r = 0;
+
+        assert(fd >= 0);
+        assert(states);
+
+        STRV_FOREACH(state, states) {
+                _cleanup_fclose_ FILE *f = NULL;
+                int k;
+
+                k = fdopen_independent(fd, "we", &f);
+                if (k < 0)
+                        return RET_GATHER(r, k);
+
+                k = write_string_stream(f, *state, WRITE_STRING_FILE_DISABLE_BUFFER);
+                if (k >= 0) {
+                        log_debug("Using sleep state '%s'.", *state);
                         return 0;
                 }
 
-                return log_debug_errno(errno, "/sys/power/resume_offset not writable: %m");
+                RET_GATHER(r, log_debug_errno(k, "Failed to write '%s' to /sys/power/state: %m", *state));
         }
 
-        xsprintf(offset_str, "%" PRIu64, hibernate_location->offset);
-        r = write_string_file("/sys/power/resume_offset", offset_str, WRITE_STRING_FILE_DISABLE_BUFFER);
-        if (r < 0)
-                return log_debug_errno(r, "Failed to write swap file offset to /sys/power/resume_offset for '%s': '%s': %m",
-                                       hibernate_location->swap->device, offset_str);
-
-        log_debug("Wrote resume_offset= value for %s to /sys/power/resume_offset: %s", hibernate_location->swap->device, offset_str);
-
-        return 0;
+        return r;
 }
 
 #if 0 /// elogind uses a special variant to heed suspension modes
-static int write_mode(char **modes) {
+static int write_mode(char * const *modes) {
         int r = 0;
 
         STRV_FOREACH(mode, modes) {
                 int k;
 
                 k = write_string_file("/sys/power/disk", *mode, WRITE_STRING_FILE_DISABLE_BUFFER);
-                if (k >= 0)
+                if (k >= 0) {
+                        log_debug("Using sleep disk mode '%s'.", *mode);
                         return 0;
+                }
 
-                log_debug_errno(k, "Failed to write '%s' to /sys/power/disk: %m", *mode);
-                if (r >= 0)
-                        r = k;
+                RET_GATHER(r, log_debug_errno(k, "Failed to write '%s' to /sys/power/disk: %m", *mode));
         }
 
         return r;
 }
+
 #else // 0
-static int write_mode(SleepOperation operation, char **modes) {
+static int write_mode(SleepOperation operation, char * const *modes) {
         int r = 0;
         static char const mode_disk[] = "/sys/power/disk";
         static char const mode_mem[] = "/sys/power/mem_sleep";
@@ -287,31 +345,6 @@ static int write_mode(SleepOperation operation, char **modes) {
 }
 #endif // 0
 
-static int write_state(FILE **f, char **states) {
-        int r = 0;
-
-        assert(f);
-        assert(*f);
-
-        STRV_FOREACH(state, states) {
-                int k;
-
-                k = write_string_stream(*f, *state, WRITE_STRING_FILE_DISABLE_BUFFER);
-                if (k >= 0)
-                        return 0;
-                log_debug_errno(k, "Failed to write '%s' to /sys/power/state: %m", *state);
-                if (r >= 0)
-                        r = k;
-
-                fclose(*f);
-                *f = fopen("/sys/power/state", "we");
-                if (!*f)
-                        return -errno;
-        }
-
-        return r;
-}
-
 #if 0 /// elogind does neither ship homed nor supports any substitution right now
 static int lock_all_homes(void) {
         _cleanup_(sd_bus_error_free) sd_bus_error error = SD_BUS_ERROR_NULL;
@@ -324,15 +357,9 @@ static int lock_all_homes(void) {
 
         r = sd_bus_open_system(&bus);
         if (r < 0)
-                return log_warning_errno(r, "Failed to connect to system bus, ignoring: %m");
+                return log_error_errno(r, "Failed to connect to system bus: %m");
 
-        r = sd_bus_message_new_method_call(
-                        bus,
-                        &m,
-                        "org.freedesktop.home1",
-                        "/org/freedesktop/home1",
-                        "org.freedesktop.home1.Manager",
-                        "LockAllHomes");
+        r = bus_message_new_method_call(bus, &m, bus_home_mgr, "LockAllHomes");
         if (r < 0)
                 return bus_log_create_error(r);
 
@@ -358,12 +385,12 @@ static int execute(
                 SleepOperation operation,
                 const char *action) {
 
-        char *arguments[] = {
+        const char *arguments[] = {
                 NULL,
-                (char*) "pre",
+                "pre",
                 /* NB: we use 'arg_operation' instead of 'operation' here, as we want to communicate the overall
                  * operation here, not the specific one, in case of s2h. */
-                (char*) sleep_operation_to_string(arg_operation),
+                sleep_operation_to_string(arg_operation),
                 NULL
         };
         static const char* const dirs[] = {
@@ -388,59 +415,71 @@ static int execute(
         log_debug_elogind("Called for '%s' (Manager is %s)", sleep_operation_to_string(operation), m ? "Set" : "NULL");
 #endif // 1
 
-        _cleanup_(hibernate_location_freep) HibernateLocation *hibernate_location = NULL;
-        _cleanup_fclose_ FILE *f = NULL;
-        char **modes, **states;
+        _cleanup_(hibernation_device_done) HibernationDevice hibernation_device = {};
+        _cleanup_close_ int state_fd = -EBADF;
+#if 1 /// If nvidia suspend is to be used, elogind has to write to its drivers proc place
+        _cleanup_close_ int driver_fd = -EBADF;
+#endif // 1
         int r;
 
         assert(sleep_config);
         assert(operation >= 0);
-        assert(operation < _SLEEP_OPERATION_MAX);
-        assert(operation != SLEEP_SUSPEND_THEN_HIBERNATE); /* Handled by execute_s2h() instead */
+        assert(operation < _SLEEP_OPERATION_CONFIG_MAX); /* Others are handled by execute_s2h() instead */
 
-        states = sleep_config->states[operation];
-        modes = sleep_config->modes[operation];
-
-        if (strv_isempty(states))
+        if (strv_isempty(sleep_config->states[operation]))
                 return log_error_errno(SYNTHETIC_ERRNO(EINVAL),
                                        "No sleep states configured for sleep operation %s, can't sleep.",
                                        sleep_operation_to_string(operation));
 
-        /* This file is opened first, so that if we hit an error,
-         * we can abort before modifying any state. */
-        f = fopen("/sys/power/state", "we");
-        if (!f)
-                return log_error_errno(errno, "Failed to open /sys/power/state: %m");
+        /* This file is opened first, so that if we hit an error, we can abort before modifying any state. */
+        state_fd = open("/sys/power/state", O_WRONLY|O_CLOEXEC);
+        if (state_fd < 0)
+                return -errno;
 
-#ifdef __GLIBC__ /// elogind must not disable buffers on musl-libc based systems
-        setvbuf(f, NULL, _IONBF, 0);
-#endif // __GLIBC__
+#if 1 /// The same with nvidia handling. elogind opens the write-to path now, so we can exit early on errors
+        if (m->handle_nvidia_sleep) {
+                /* This file is opened first, so that if we hit an error, we can abort before modifying any state. */
+                driver_fd = open("/proc/driver/nvidia/suspend", O_WRONLY | O_CLOEXEC);
+                if (driver_fd < 0)
+                        return -errno;
+        }
+#endif // 1
 
         /* Configure hibernation settings if we are supposed to hibernate */
 #if 0 /// elogind supports suspend modes, and keeps its config, so checking modes for emptiness alone doesn't cut it
-        if (!strv_isempty(modes)) {
+        if (sleep_operation_is_hibernation(operation)) {
 #else // 0
         if (operation != SLEEP_SUSPEND) {
 #endif // 0
-                r = find_hibernate_location(&hibernate_location);
+                bool resume_set;
+
+                r = find_suitable_hibernation_device(&hibernation_device);
                 if (r < 0)
                         return log_error_errno(r, "Failed to find location to hibernate to: %m");
-                if (r == 0) { /* 0 means: no hibernation location was configured in the kernel so far, let's
-                               * do it ourselves then. > 0 means: kernel already had a configured hibernation
-                               * location which we shouldn't touch. */
-                        r = write_hibernate_location_info(hibernate_location);
+                resume_set = r > 0;
+
+                r = write_efi_hibernate_location(&hibernation_device, !resume_set);
+                if (!resume_set) {
+                        if (r == -EOPNOTSUPP)
+                                return log_error_errno(r, "No valid 'resume=' option found, refusing to hibernate.");
                         if (r < 0)
-                                return log_error_errno(r, "Failed to prepare for hibernation: %m");
+                                return r;
+
+                        r = write_resume_config(hibernation_device.devno, hibernation_device.offset, hibernation_device.path);
+                        if (r < 0)
+                                goto fail;
                 }
 
 #if 0 /// elogind supports suspend modes, and our write_mode() variant does more and logs on error
-                r = write_mode(modes);
-                if (r < 0)
-                        return log_error_errno(r, "Failed to write mode to /sys/power/disk: %m");
+                r = write_mode(sleep_config->modes[operation]);
+                if (r < 0) {
+                        log_error_errno(r, "Failed to write mode to /sys/power/disk: %m");
+                        goto fail;
+                }
         }
 #else // 0
         }
-        (void)write_mode(operation, modes);
+        (void)write_mode(operation, sleep_config->modes[operation]);
 #endif // 0
 
         /* Pass an action string to the call-outs. This is mostly our operation string, except if the
@@ -448,12 +487,11 @@ static int execute(
         if (!action)
                 action = sleep_operation_to_string(operation);
 
-        r = setenv("SYSTEMD_SLEEP_ACTION", action, 1);
-        if (r != 0)
-                log_warning_errno(errno, "Error setting SYSTEMD_SLEEP_ACTION=%s, ignoring: %m", action);
+        if (setenv("SYSTEMD_SLEEP_ACTION", action, /* overwrite = */ 1) < 0)
+                log_warning_errno(errno, "Failed to set SYSTEMD_SLEEP_ACTION=%s, ignoring: %m", action);
 
 #if 0 /// elogind allows admins to configure that hook scripts must succeed. The systemd default does not cut it here
-        (void) execute_directories(dirs, DEFAULT_TIMEOUT_USEC, NULL, NULL, arguments, NULL, EXEC_DIR_PARALLEL | EXEC_DIR_IGNORE_ERRORS);
+        (void) execute_directories(dirs, DEFAULT_TIMEOUT_USEC, NULL, NULL, (char **) arguments, NULL, EXEC_DIR_PARALLEL | EXEC_DIR_IGNORE_ERRORS);
         (void) lock_all_homes();
 #else // 0
         m->callback_failed = false;
@@ -462,7 +500,7 @@ static int execute(
         log_debug_elogind("Executing suspend hook scripts... (Must succeed: %s)",
                           m->callback_must_succeed ? "YES" : "no");
 
-        r = execute_directories(dirs, DEFAULT_TIMEOUT_USEC, gather_output, gather_args, arguments, NULL, EXEC_DIR_NONE);
+        r = execute_directories(dirs, DEFAULT_TIMEOUT_USEC, gather_output, gather_args, (char **) arguments, NULL, EXEC_DIR_NONE);
 
         log_debug_elogind("Result is %d (callback_failed: %s)", r, m->callback_failed ? "true" : "false");
 
@@ -475,7 +513,7 @@ static int execute(
                 }
 
                 if ( m->broadcast_suspend_interrupts )
-                        utmp_wall(l, "root", "n/a", logind_wall_tty_filter, m);
+                        wall(l, "root", "n/a", logind_wall_tty_filter, m);
 
                 log_error_errno(r, "MESSAGE_ID=" SD_MESSAGE_SLEEP_STOP_STR " %s", l);
 
@@ -495,20 +533,20 @@ static int execute(
 
         log_struct(LOG_INFO,
                    "MESSAGE_ID=" SD_MESSAGE_SLEEP_START_STR,
-                   LOG_MESSAGE("Entering sleep state '%s'...", sleep_operation_to_string(operation)),
+                   LOG_MESSAGE("Performing sleep operation '%s'...", sleep_operation_to_string(operation)),
                    "SLEEP=%s", sleep_operation_to_string(arg_operation));
 
 #if 1 /// elogind may try to send a suspend signal to an nvidia card
         if ( m->handle_nvidia_sleep )
-                have_nvidia = nvidia_sleep(m, operation, &vtnr);
+                have_nvidia = nvidia_sleep(m, driver_fd, operation, &vtnr);
 #endif // 1
 
 #if 0 /// Instead of only writing to /sys/power/state, elogind offers the possibility to call an extra program instead
-        r = write_state(&f, states);
+        r = write_state(state_fd, sleep_config->states[operation]);
 #else // 0
         r = execute_external(m, operation);
         if (r < 0)
-                r = write_state(&f, states);
+                r = write_state(state_fd, sleep_config->states[operation]);
 #endif // 0
         if (r < 0)
                 log_struct_errno(LOG_ERR, r,
@@ -518,24 +556,68 @@ static int execute(
         else
                 log_struct(LOG_INFO,
                            "MESSAGE_ID=" SD_MESSAGE_SLEEP_STOP_STR,
-                           LOG_MESSAGE("System returned from sleep state."),
+                           LOG_MESSAGE("System returned from sleep operation '%s'.", sleep_operation_to_string(arg_operation)),
                            "SLEEP=%s", sleep_operation_to_string(arg_operation));
 
 #if 1 /// if put to sleep, elogind also has to wakeup an nvidia card
         if (have_nvidia)
-                nvidia_sleep(m, _SLEEP_OPERATION_MAX, &vtnr);
+                nvidia_sleep(m, driver_fd, _SLEEP_OPERATION_MAX, &vtnr);
 #endif // 1
 
-        arguments[1] = (char*) "post";
+        arguments[1] = "post";
 #if 0 /// elogind does not execute wakeup hook scripts in parallel, they might be order relevant
-        (void) execute_directories(dirs, DEFAULT_TIMEOUT_USEC, NULL, NULL, arguments, NULL, EXEC_DIR_PARALLEL | EXEC_DIR_IGNORE_ERRORS);
+        (void) execute_directories(dirs, DEFAULT_TIMEOUT_USEC, NULL, NULL, (char **) arguments, NULL, EXEC_DIR_PARALLEL | EXEC_DIR_IGNORE_ERRORS);
 #else // 0
-        (void) execute_directories(dirs, DEFAULT_TIMEOUT_USEC, NULL, NULL, arguments, NULL, EXEC_DIR_IGNORE_ERRORS);
+        (void) execute_directories(dirs, DEFAULT_TIMEOUT_USEC, NULL, NULL, (char **) arguments, NULL, EXEC_DIR_IGNORE_ERRORS);
 #endif // 0
 
+        if (r >= 0)
+                return 0;
+
+fail:
+        if (sleep_operation_is_hibernation(operation) && is_efi_boot())
+                (void) efi_set_variable(EFI_SYSTEMD_VARIABLE(HibernateLocation), NULL, 0);
+
         return r;
 }
 
+/* Return true if wakeup type is APM timer */
+static int check_wakeup_type(void) {
+        static const char dmi_object_path[] = "/sys/firmware/dmi/entries/1-0/raw";
+        uint8_t wakeup_type_byte, tablesize;
+        _cleanup_free_ char *buf = NULL;
+        size_t bufsize;
+        int r;
+
+        /* implementation via dmi/entries */
+        r = read_full_virtual_file(dmi_object_path, &buf, &bufsize);
+        if (r < 0)
+                return log_debug_errno(r, "Unable to read %s: %m", dmi_object_path);
+        if (bufsize < 25)
+                return log_debug_errno(SYNTHETIC_ERRNO(EINVAL),
+                                       "Only read %zu bytes from %s (expected 25)",
+                                       bufsize, dmi_object_path);
+
+        /* index 1 stores the size of table */
+        tablesize = (uint8_t) buf[1];
+        if (tablesize < 25)
+                return log_debug_errno(SYNTHETIC_ERRNO(EINVAL),
+                                       "Table size less than the index[0x18] where waketype byte is available.");
+
+        wakeup_type_byte = (uint8_t) buf[24];
+        /* 0 is Reserved and 8 is AC Power Restored. As per table 12 in
+         * https://www.dmtf.org/sites/default/files/standards/documents/DSP0134_3.4.0.pdf */
+        if (wakeup_type_byte >= 128)
+                return log_debug_errno(SYNTHETIC_ERRNO(EINVAL), "Expected value in range 0-127");
+
+        if (wakeup_type_byte == 3) {
+                log_debug("DMI BIOS System Information indicates wakeup type is APM Timer");
+                return true;
+        }
+
+        return false;
+}
+
 static int custom_timer_suspend(const SleepConfig *sleep_config) {
         usec_t hibernate_timestamp;
         int r;
@@ -546,7 +628,7 @@ static int custom_timer_suspend(const SleepConfig *sleep_config) {
 
         while (battery_is_discharging_and_low() == 0) {
                 _cleanup_hashmap_free_ Hashmap *last_capacity = NULL, *current_capacity = NULL;
-                _cleanup_close_ int tfd = -1;
+                _cleanup_close_ int tfd = -EBADF;
                 struct itimerspec ts = {};
                 usec_t suspend_interval;
                 bool woken_by_timer;
@@ -654,8 +736,7 @@ static int freeze_thaw_user_slice(const char **method) {
         if (r < 0)
                 return log_debug_errno(r, "Failed to open connection to systemd: %m");
 
-        /* Wait for 1.5 seconds at maximum for freeze operation */
-        (void) sd_bus_set_method_call_timeout(bus, 1500 * USEC_PER_MSEC);
+        (void) sd_bus_set_method_call_timeout(bus, FREEZE_TIMEOUT);
 
         r = bus_call_method(bus, bus_systemd_mgr, *method, &error, NULL, "s", SPECIAL_USER_SLICE);
         if (r < 0)
@@ -747,7 +828,8 @@ static int help(void) {
                "  hibernate              Hibernate the system\n"
                "  hybrid-sleep           Both hibernate and suspend the system\n"
                "  suspend-then-hibernate Initially suspend and then hibernate\n"
-               "                         the system after a fixed period of time\n"
+               "                         the system after a fixed period of time or\n"
+               "                         when battery is low\n"
                "\nSee the %s for details.\n",
                program_invocation_short_name,
                link);
@@ -756,6 +838,7 @@ static int help(void) {
 }
 
 static int parse_argv(int argc, char *argv[]) {
+
         enum {
                 ARG_VERSION = 0x100,
         };
@@ -773,6 +856,7 @@ static int parse_argv(int argc, char *argv[]) {
 
         while ((c = getopt_long(argc, argv, "h", options, NULL)) >= 0)
                 switch (c) {
+
                 case 'h':
                         return help();
 
@@ -784,6 +868,7 @@ static int parse_argv(int argc, char *argv[]) {
 
                 default:
                         assert_not_reached();
+
                 }
 
         if (argc - optind != 1)
@@ -799,7 +884,7 @@ static int parse_argv(int argc, char *argv[]) {
 }
 
 static int run(int argc, char *argv[]) {
-        _cleanup_(free_sleep_configp) SleepConfig *sleep_config = NULL;
+        _cleanup_(sleep_config_freep) SleepConfig *sleep_config = NULL;
         int r;
 
         log_setup();
@@ -841,7 +926,7 @@ int do_sleep(Manager *sleep_config, SleepOperation operation) {
                          * asked us to do both: suspend + hibernate, and it's almost certainly the
                          * hibernation that failed, hence still do the other thing, the suspend. */
 
-                        log_notice("Couldn't hybrid sleep, will try to suspend instead.");
+                        log_notice_errno(r, "Couldn't hybrid sleep, will try to suspend instead: %m");
 
                         r = execute(sleep_config, SLEEP_SUSPEND, "suspend-after-failed-hybrid-sleep");
                 }
@@ -851,6 +936,7 @@ int do_sleep(Manager *sleep_config, SleepOperation operation) {
         default:
                 r = execute(sleep_config, arg_operation, NULL);
                 break;
+
         }
 
         return r;
